%module transferfunction

%{
#include "libharkio3.h"
#include "libharkio3_macros.h"
#define HARKIO_LOG_INITIAL_LEVEL harkio_Log_Error

%}

typedef struct{
    struct zip *handle;         /*!< 伝達関数ファイルのハンドル */
    harkio_Positions *poses;    /*!< TSP測定位置を表すメンバ */
    harkio_Positions *mics;     /*!< マイク位置を表すメンバ */
    harkio_Config *cfg;         /*!< 一般設定を表すメンバ */
    harkio_Neighbors *nbrs;     /*!< TSP測定位置の近接関係を表すメンバ */
    harkio_Matrix **loctfs;     /*!< 定位用伝達関数を表すメンバ */
    harkio_Matrix **septfs;     /*!< 分離用伝達関数を表すメンバ */
} harkio_TransferFunction;

harkio_TransferFunction* harkio_TransferFunction_new(void);
harkio_TransferFunction* harkio_TransferFunction_fromFile(char *path, char *signature);
harkio_TransferFunction* harkio_TransferFunction_fromData(harkio_Positions *poses, harkio_Positions *mics, harkio_Config *cfg, harkio_Neighbors *nbrs);
void harkio_TransferFunction_delete(harkio_TransferFunction **tf);

int harkio_TransferFunction_write(harkio_TransferFunction *tf, char *path, char *signature);
// int harkio_TransferFunction_readTFs(harkio_TransferFunction *tf);

harkio_Config* harkio_TransferFunction_getConfig(harkio_TransferFunction *tf);
harkio_Neighbors* harkio_TransferFunction_getNeighbors(harkio_TransferFunction *tf);
harkio_Positions* harkio_TransferFunction_getMicrophones(harkio_TransferFunction *tf);
harkio_Positions* harkio_TransferFunction_getPositions(harkio_TransferFunction *tf);

int harkio_TransferFunction_setConfig(harkio_TransferFunction *tf, harkio_Config *value);
int harkio_TransferFunction_setNeighbors(harkio_TransferFunction *tf, harkio_Neighbors *value);
int harkio_TransferFunction_setMicrophones(harkio_TransferFunction *tf, harkio_Positions *value);
int harkio_TransferFunction_setPositions(harkio_TransferFunction *tf, harkio_Positions *value);

harkio_Matrix* harkio_TransferFunction_getTFByPos(harkio_TransferFunction *tf, harkio_Position *pos, float tolerance, enum TransferType tf_type);
harkio_Matrix* harkio_TransferFunction_getTFById(harkio_TransferFunction *tf, int id, enum TransferType tf_type);
int harkio_TransferFunction_setLocTFs(harkio_TransferFunction *tf, harkio_Matrix **tfs);
int harkio_TransferFunction_setSepTFs(harkio_TransferFunction *tf, harkio_Matrix **tfs);



//int libharkio3_TransferFunction_readlibhakio2(harkio_TransferFunction **tf, char *loctfpath , char *septfpath);
//int libharkio3_SeparationMatrix_readlibhakio2(harkio_TransferFunction **tf, char *sepmatpath , char *tfzippath);
harkio_TransferFunction* harkio_Compat_TransferFunction_readlibharkio2(char *loctfpath , char *septfpath);
harkio_TransferFunction* harkio_Compat_SeparationMatrix_readlibharkio2(char *sepmatpath , char *tfzippath);
